/*
** $Id: smrazor.c,v 1.24 2002/08/01 20:23:51 peterh Exp $
*/
#include <config.h>

#include <errno.h>
#include <limits.h>
#include <signal.h>
#include <stdarg.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include "libmilter/mfapi.h"

#ifndef true
typedef int bool;
# define false	0
# define true	1
#endif /* ! true */

#ifdef DEBUG
#define DPRINTF(x) dprintf##x
#else /* DEBUG */
#define DPRINTF(x)
#endif /* DEBUG */

#define PERL_BIN_PATH "/usr/bin/perl"
#define RAZOR_CHECK_PATH "/usr/local/bin/razor-check"
#define DEFAULT_PID_FILE_NAME "/var/run/smrazor.pid"
#define ERROR_FILE_NAME "/tmp/smrazor.err"
#define ERR_TIME_FORMAT "[%m/%d/%Y %H:%M:%S] "
#define RAZOR_CHECK_TIMEOUT 30

struct mlfiPriv
{
	FILE* read_pipe;
	FILE* write_pipe;
	int razor_pid;
	int on_cr;
	char* from;
	char* subject;
};

#define MAX_CONNECTION_STRING_LEN 1024
#define MAX_ERROR_STR_LEN 4096

#define MLFIPRIV	((struct mlfiPriv *) smfi_getpriv(ctx))
#define NULL_STR(str)	((str == NULL ? "" : str))

extern void mlfi_cleanup(SMFICTX *);

void
dprintf(const char* format, ...)
{
	FILE* error_file;
	va_list ap;
	time_t current_clock;
	struct tm* current_time;
	char time_str[MAX_ERROR_STR_LEN];

	error_file = fopen(ERROR_FILE_NAME, "a");
	if (error_file == NULL)
	{
		fprintf(stderr, "Error opening error file (%s): %s\n", ERROR_FILE_NAME, strerror(errno));
	}
	current_clock = time(NULL);
	current_time = localtime(&current_clock);
	strftime(time_str, MAX_ERROR_STR_LEN, ERR_TIME_FORMAT, current_time);
	fprintf(error_file, "%s(%d) ", time_str, getpid());
	va_start(ap, format);
	vfprintf(error_file, format, ap);
	va_end(ap);
	fclose(error_file);
}

int
razor_popen(priv)
	struct mlfiPriv* priv;
{
	char* argv[5];
	int pdes[2];
	int fdes[2];
	int pid;
	FILE* error_file;

	argv[0] = PERL_BIN_PATH;
	argv[1] = RAZOR_CHECK_PATH;
#ifdef DEBUG
	argv[2] = "-d";
	argv[3] = NULL;
#else /* DEBUG */
	argv[2] = NULL;
#endif /* DEBUG */

	if (pipe(pdes) < 0)
		return -1;

	if (pipe(fdes) < 0)
		return -1;

	switch (pid = vfork())
	{
		case -1:
			close(pdes[0]);
			close(pdes[1]);
			return -1;
		case 0:
			if (pdes[0] != STDIN_FILENO)
			{
				dup2(pdes[0], STDIN_FILENO);
				close(pdes[0]);
			}
			if (fdes[1] != STDOUT_FILENO)
			{
				dup2(fdes[1], STDOUT_FILENO);
				close(fdes[1]);
			}
			close(pdes[1]);
			close(fdes[0]);
			execv(PERL_BIN_PATH, argv);
			DPRINTF(("Error creating pipe: %s\n", strerror(errno)));
			_exit(127);
		default:
			DPRINTF(("launched razor-check(%d)\n", pid));
	}
	priv->write_pipe = fdopen(pdes[1], "w");
	priv->read_pipe = fdopen(fdes[0], "r");
	priv->razor_pid = pid;
	close(pdes[0]);
	close(fdes[1]);

	return 0;
}

int
razor_pclose(priv)
	struct mlfiPriv* priv;
{
	int pid;
	int pstat;

	do
	{
		pid = wait4(priv->razor_pid, &pstat, 0, (struct rusage*)0);
	} while (pid == -1 && errno == EINTR);

	priv->razor_pid = 0;

	return (pid == -1 ? -1 : pstat);
}

sfsistat
mlfi_envfrom(ctx, envfrom)
	SMFICTX *ctx;
	char **envfrom;
{
	struct mlfiPriv *priv;
	int fd = -1;

	DPRINTF(("mlfi_envfrom(%p);\n", ctx));

	/* allocate some private memory */
	priv = malloc(sizeof *priv);
	if (priv == NULL)
	{
		DPRINTF(("Error alloctating memory for private data.\n"));
		/* can't accept this message right now */
		return SMFIS_TEMPFAIL;
	}
	memset(priv, '\0', sizeof *priv);

	/* open a file to store this message */
	if (razor_popen(priv) != 0)
	{
		/* can't accept this message right now */
		DPRINTF(("Error opening pipe to razor.\n"));
		free(priv);
		return SMFIS_TEMPFAIL;
	}
		
	/* save the private data */
	smfi_setpriv(ctx, priv);

	/* continue processing */
	return SMFIS_CONTINUE;
}

sfsistat
mlfi_header(ctx, headerf, headerv)
	SMFICTX *ctx;
	char *headerf;
	char *headerv;
{
//	DPRINTF(("mlfi_header(%p, %s, %s)\n", ctx, headerf, headerv));
	if (strncasecmp("From", headerf, 4) == 0)
	{
		MLFIPRIV->from = strdup(headerv);
	}
	if (strncasecmp("Subject", headerf, 7) == 0)
	{
		MLFIPRIV->subject = strdup(headerv);
	}
	if (strncasecmp("To", headerf, 2) == 0)
	{
		if (strncasecmp("spam", headerv, 4) == 0)
		{
			return SMFIS_ACCEPT;
		}
	}
	fprintf(MLFIPRIV->write_pipe, "%s: %s\n", headerf, headerv);

	/* continue processing */
	return SMFIS_CONTINUE;
}

sfsistat
mlfi_eoh(ctx)
	SMFICTX *ctx;
{
	DPRINTF(("mlfi_eoh(%p)\n", ctx));
	/* output the blank line between the header and the body */
	fprintf(MLFIPRIV->write_pipe, "\n");

	/* continue processing */
	return SMFIS_CONTINUE;
}

char*
find_char(str, target_char, len)
	u_char* str;
	u_char target_char;
	size_t len;
{
	int i;

	for (i = 0; i < len; ++i)
	{
		if (str[i] == target_char)
		{
			return str + i;
		}
	}

	return NULL;
}


sfsistat
mlfi_body(ctx, bodyp, bodylen)
	SMFICTX *ctx;
	u_char *bodyp;
	size_t bodylen;
{
	/* output body block to log file */
	struct mlfiPriv* priv = MLFIPRIV;
	u_char* cr_pos;
	u_char* last_cr_pos = bodyp;
	size_t remaining = bodylen;
	int result;

	DPRINTF(("mlfi_body(%p)\n", ctx));

	if (bodylen <= 0)
	{
		DPRINTF(("Unexpected body length: %d\n", bodylen));
		mlfi_cleanup(ctx);
		return SMFIS_TEMPFAIL;
	}

	if (priv->on_cr)
	{
		if (bodyp[0] != '\n')
		{
			result = fwrite("\r", 1, 1, priv->write_pipe);
			if (result != 1)
			{
				DPRINTF(("Error writing to razor pipe: %s\n", strerror(errno)));
				DPRINTF(("1: 1 byte\n"));
				mlfi_cleanup(ctx);
				return SMFIS_TEMPFAIL;
			}
		}
		priv->on_cr = 0;
	}

	cr_pos = find_char(bodyp, '\r', bodylen);
	while (cr_pos != NULL)
	{
		if (cr_pos + 1 >= bodyp + bodylen)
		{
			result = fwrite(last_cr_pos, cr_pos - last_cr_pos,
					1, priv->write_pipe);
			if (result <= 0)
			{
				DPRINTF(("Error writing to razor pipe: %s\n", strerror(errno)));
				DPRINTF(("2: last_cr_pos:%p cr_pos: %p diff: %d\n", last_cr_pos, cr_pos, cr_pos - last_cr_pos));

				mlfi_cleanup(ctx);
				return SMFIS_TEMPFAIL;
			}
			priv->on_cr = 1;
			last_cr_pos = bodyp + bodylen;
			break;
		} 

		if (cr_pos == last_cr_pos)
		{
		}
		else if (*(cr_pos + 1) == '\n')
		{
			result = fwrite(last_cr_pos, cr_pos - last_cr_pos, 
					1, priv->write_pipe);
			if (result <= 0)
			{
				DPRINTF(("Error writing to razor pipe: %s\n", strerror(errno)));
				DPRINTF(("3: last_cr_pos:%p cr_pos: %p diff: %d\n", last_cr_pos, cr_pos, cr_pos - last_cr_pos));
				mlfi_cleanup(ctx);
				return SMFIS_TEMPFAIL;
			}
		}
		else
		{
			result = fwrite(last_cr_pos, cr_pos - last_cr_pos + 1, 
					1, priv->write_pipe);
			if (result <= 0)
			{
				DPRINTF(("Error writing to razor pipe: %s\n", strerror(errno)));
				DPRINTF(("4: last_cr_pos:%p cr_pos: %p diff: %d\n", last_cr_pos, cr_pos, cr_pos - last_cr_pos + 1));
				mlfi_cleanup(ctx);
				return SMFIS_TEMPFAIL;
			}
		}
		last_cr_pos = cr_pos + 1;
		if (last_cr_pos == bodyp + bodylen)
		{
			cr_pos = NULL;
		}
		else
		{
			cr_pos = find_char(last_cr_pos, '\r', 
					   bodyp + bodylen - last_cr_pos);
		}
	}
	if (last_cr_pos != bodyp + bodylen)
	{
		result = fwrite(last_cr_pos, bodyp + bodylen - last_cr_pos, 1,
			        priv->write_pipe);
		if (result <= 0)
		{
			DPRINTF(("Error writing to razor pipe: %s\n", strerror(errno)));
			DPRINTF(("5: last_cr_pos:%p bodyp:%p bodylen:%d calc:%d\n", last_cr_pos, cr_pos, bodyp + bodylen - last_cr_pos));

			mlfi_cleanup(ctx);
			return SMFIS_TEMPFAIL;
		}
	}

	/* continue processing */
	return SMFIS_CONTINUE;
}
#define BUF_LEN 4096

sfsistat
mlfi_eom(ctx)
	SMFICTX *ctx;
{
	sfsistat rstat = SMFIS_CONTINUE;
	struct mlfiPriv *priv = MLFIPRIV;
	int result;
	FILE* error_file;
	FILE* log_file;
	char buf[BUF_LEN];
	int read_fd;
	fd_set read_fds;
	struct timeval select_timeout;

	DPRINTF(("mlfi_eom(%p)\n", ctx));

	if (priv->write_pipe == NULL)
	{
		DPRINTF(("mlfi_eom: razor pipe is null.\n"));
		mlfi_cleanup(ctx);
		return SMFIS_TEMPFAIL;
	}

	result = fclose(priv->write_pipe);
	priv->write_pipe = NULL;
	if (result == 0)
	{
		error_file = fopen(ERROR_FILE_NAME, "a");
		read_fd = fileno(priv->read_pipe);
		FD_ZERO(&read_fds);
		FD_SET(read_fd, &read_fds);
		select_timeout.tv_sec = RAZOR_CHECK_TIMEOUT;
		select_timeout.tv_usec = 0;
		DPRINTF(("Starting select (%p).\n", ctx));
		result = select(read_fd + 1, &read_fds, NULL, NULL, 
				&select_timeout);
		DPRINTF(("Select done (%p): %d %d\n", ctx, result, errno));
		if (result == 1)
		{
			while ((result = fread(buf, 1, BUF_LEN, 
					       priv->read_pipe)) > 0)
			{
				buf[result] = 0;
				fprintf(error_file, buf);
			}
		}
		else
		{
			kill(priv->razor_pid, SIGKILL);
			dprintf("Error reading from razor-check (%p) (%d): timeout\n", ctx, priv->razor_pid);
		}
		fclose(error_file);
		fclose(priv->read_pipe);
		priv->read_pipe = NULL;
		result = razor_pclose(priv);
		if (WIFEXITED(result))
		{
			if (WEXITSTATUS(result) != 0)
			{
				smfi_addheader(ctx, "X-SMRazor", "ok");
				DPRINTF(("Accepted: %p\n", ctx));
			}
			else
			{
				smfi_addheader(ctx, "X-SMRazor", "SPAM");
				DPRINTF(("Rejected: %p\n", ctx));
				log_file = fopen("/tmp/razored.log", "a");
				if (log_file != NULL)
				{
					fprintf(log_file, "%s\t%s\n", 
						NULL_STR(priv->from),
						NULL_STR(priv->subject));
				}
				fclose(log_file);
			}
		}
		else
		{
			DPRINTF(("Bad exit status from razor pipe: %d\n", result));
			rstat = SMFIS_TEMPFAIL;
		}
	}

	mlfi_cleanup(ctx);

	return rstat;
}

sfsistat
mlfi_close(ctx)
	SMFICTX *ctx;
{
	DPRINTF(("mlfi_close(%p)\n", ctx));
	return SMFIS_ACCEPT;
}

sfsistat
mlfi_abort(ctx)
	SMFICTX *ctx;
{
	DPRINTF(("mlfi_abort(%p)\n", ctx));
	mlfi_cleanup(ctx);
	return SMFIS_CONTINUE;
}

void
mlfi_cleanup(ctx)
	SMFICTX *ctx;
{
	struct mlfiPriv *priv = MLFIPRIV;
	int result;

	DPRINTF(("mlfi_cleanup(%p)\n", ctx));

	if (priv == NULL)
		return;

	if (priv->read_pipe != NULL)
	{
		fclose(priv->read_pipe);
	}
	if (priv->write_pipe != NULL)
	{
		fclose(priv->write_pipe);
	}
	if (priv->razor_pid != 0)
	{
		kill(priv->razor_pid, SIGKILL);
		result = razor_pclose(priv);
		if (!WIFEXITED(result))
		{
			DPRINTF(("Bad exit status from razor pclose: %d\n", 
				 result));
		}
	}

	/* release private memory */
	free(priv);
	smfi_setpriv(ctx, NULL);

	return;
}

struct smfiDesc smfilter =
{
	"SMRazor",	/* filter name */
	SMFI_VERSION,	/* version code -- do not change */
	SMFIF_ADDHDRS,	/* flags */
	NULL,		/* connection info filter */
	NULL,		/* SMTP HELO command filter */
	mlfi_envfrom,	/* envelope sender filter */
	NULL,		/* envelope recipient filter */
	mlfi_header,	/* header filter */
	mlfi_eoh,	/* end of header */
	mlfi_body,	/* body block filter */
	mlfi_eom,	/* end of message */
	mlfi_abort,	/* message aborted */
	mlfi_close	/* connection cleanup */
};

#define MAX_CONNECTION_STRING_LEN 1024

int
main(argc, argv)
	int argc;
	char *argv[];
{
	int c;
	int result;
	const char *args = "p:c:";
	char connection_string[MAX_CONNECTION_STRING_LEN];
	char pid_file_name[PATH_MAX];
	FILE* pid_file;

	connection_string[0] = 0;
	strncpy(pid_file_name, DEFAULT_PID_FILE_NAME, PATH_MAX);

	/* Process command line options */
	while ((c = getopt(argc, argv, args)) != -1)
	{
		switch (c)
		{
	  	  case 'c':
			if (optarg == NULL || *optarg == '\0')
			{
				(void) fprintf(stderr, "Illegal conn: %s\n",
					       optarg);
				exit(EX_USAGE);
			}
			strncpy(pid_file_name, optarg, PATH_MAX);
			break;

		  case 'p':
			if (optarg == NULL || *optarg == '\0')
			{
				(void) fprintf(stderr, "Illegal conn: %s\n",
					       optarg);
				exit(EX_USAGE);
			}
			strncpy(connection_string, optarg, 
				MAX_CONNECTION_STRING_LEN);
			(void) smfi_setconn(optarg);
			break;

		}
	}

	pid_file = fopen(pid_file_name, "w");
	if (pid_file == NULL)
	{
		fprintf(stderr, "Error opening pid file (%s): %s\n", pid_file_name, strerror(errno));
		exit(EX_CANTCREAT);
	}
	result = fprintf(pid_file, "%d", getpid());
	if (result < 1)
	{
		fprintf(stderr, "Error writting to pid file (%s): %s\n", pid_file_name, strerror(errno));
		exit(EX_IOERR);
	}
	result = fclose(pid_file);
	if (result != 0)
	{
		fprintf(stderr, "Error closing pid file (%s): %s\n", pid_file_name, strerror(errno));
		exit(EX_IOERR);
	}
	
	if (smfi_register(smfilter) == MI_FAILURE)
	{
		fprintf(stderr, "smfi_register failed\n");
		exit(EX_UNAVAILABLE);
	}
	result = smfi_main();
	if (result != 0)
	{
		fprintf(stderr, "Error running milter. Check mail log for details.\n");
	}
	if (strncasecmp("local:", connection_string, 6) == 0)
	{
		unlink(connection_string + 6);
	}

	return result;
}

/* eof */

